(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: LightingVS.cpp
//
// Desc: Example code showing how to use D3D fixed-function lights and vertex 
//       shader lights.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
unit LightingVSUnit;

interface

uses
  Windows, Messages,
  Direct3D9, D3DX9,
  D3DApp, D3DFont, D3DUtil, DXUtil,
  Resource;

{.$DEFINE DEBUG_VS}   // Uncomment this line to debug vertex shaders
{.$DEFINE DEBUG_PS}   // Uncomment this line to debug pixel shaders

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------

type
  // Custom D3D vertex format used by the vertex buffer
  PMyVertex = ^TMyVertex;
  TMyVertex = packed record
    p: TD3DXVector3;       // vertex position
    n: TD3DXVector3;       // vertex normal
  // static const DWORD FVF;
  end;
  
  PMyVertexArray = ^TMyVertexArray;
  TMyVertexArray = array [0..0] of TMyVertex;

const
  TMyVertex_FVF = D3DFVF_XYZ or D3DFVF_NORMAL;

const
  MAXLIGHTS{DWORD} = 7;

  // Number of vectors per light in vertex shader constants
  NUM_VECTORS_PER_LIGHT{Integer} = 7;

type
//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the
//       generic functionality needed in all Direct3D samples. CMyD3DApplication
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
    m_pWallMesh: ID3DXMesh;            // Tessellated plane to serve as the walls and floor
    m_pSphereMesh: ID3DXMesh;          // Representation of a point light
    m_pConeMesh: ID3DXMesh;            // Representation of a dir/spot light
    m_pFont: CD3DFont;                 // Font for drawing text
    m_pFontSmall: CD3DFont;            // Font for drawing text
    m_light: array[0..MAXLIGHTS-1] of TD3DLight9; // Description of the D3D lights
    m_n: LongWord;                     // Number of vertices in the wall mesh along X
    m_m: LongWord;                     // Number of vertices in the wall mesh along Z
    m_nTriangles: LongWord;            // Number of triangles in the wall mesh
    m_bShowHelp: Boolean;
    m_bWireframe: Boolean;

  private
    procedure SetMatrices(const pMatWorld, pMatViewProj: TD3DXMatrix);
    procedure SetLight(LightIndex: DWORD; const light: TD3DLight9);
  protected
    function BuildWallMesh: HRESULT;

    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function FrameMove: HRESULT; override;
    function Render: HRESULT; override;
    function FinalCleanup: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

  public
    m_dwNumLights:        DWORD;          // Number of moving lights
    m_pEffect:            ID3DXEffect;    // Effect containing vertex shaders
    m_iCurTechnique:      LongWord;       // Curent technique index
    m_pCurTechniqueName:  PAnsiChar;      // Name of current technique
    m_bUseVertexShaders:  Boolean;        // Whether to use shaders or FF
    m_bTechniqueValid:    Boolean;        // Whether current technique is supported
    m_matProj:            TD3DXMatrixA16; // Current projection matrix
    m_matView:            TD3DXMatrixA16; // Current view matrix
    m_light0:             TD3DLight9;     // Light used to add ambient lighting to the walls
    m_light1:             TD3DLight9;     // Light which affects light objects
    m_CurrentLightType:   TD3DLightType;  // Current type of moving lights (0 to cycle though light types)
  public
    constructor Create;

    function MsgProc(hWnd: HWND; uMsg: LongWord; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;



implementation

uses SysUtils;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  m_strWindowTitle            := 'Lighting';
  m_d3dEnumeration.AppUsesDepthBuffer := False;

  m_pFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall := CD3DFont.Create('Arial', 9, D3DFONT_BOLD);
  m_pWallMesh                 := nil;
  m_pSphereMesh               := nil;
  m_pConeMesh                 := nil;
  m_pEffect                   := nil;
  m_iCurTechnique             := 0;
  m_pCurTechniqueName         := nil;

  // Set up wall/floor mesh resolution.  Try changing m_n and m_m to see
  // how the lighting is affected.
  m_n := 32;
  m_m := 32;
  m_nTriangles := (m_n-1)*(m_m-1)*2;
  m_bShowHelp := False;
  m_bWireframe := False;
  m_dwNumLights := 2;
  m_bUseVertexShaders := True;
  m_CurrentLightType := TD3DLightType(0);   // Cycle through light types
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Debugging vertex shaders requires either REF or software vertex processing
  // and debugging pixel shaders requires REF.
  {$IFDEF DEBUG_VS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF) and
       ((dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING) = 0)
    then Exit;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    if (pCaps.DeviceType <> D3DDEVTYPE_REF)
    then Exit;
  {$ENDIF}

  // The sample needs at least vertex shader 2_0. Software vertex processing
  // supports vs_3_0, so it is allowed.
  if (pCaps.VertexShaderVersion >= D3DVS_VERSION(2,0)) or
     (dwBehavior and D3DCREATE_SOFTWARE_VERTEXPROCESSING <> 0) then
  begin
    Result:= S_OK;
    Exit;
  end;

  Result:= E_FAIL;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
var
  dwShaderFlags: DWORD;
  strPath: array[0..MAX_PATH-1] of AnsiChar;
  TechDesc: TD3DXTechniqueDesc;
  hTech: TD3DXHandle;
begin
  Result := m_pFont.InitDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  Result := m_pFontSmall.InitDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  // Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders with the shader debugger.
  // Debugging vertex shaders requires either REF or software vertex processing, and debugging
  // pixel shaders requires REF.  The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug
  // experience in the shader debugger.  It enables source level debugging, prevents instruction
  // reordering, prevents dead code elimination, and forces the compiler to compile against the next
  // higher available software target, which ensures that the unoptimized shaders do not exceed
  // the shader model limitations.  Setting these flags will cause slower rendering since the shaders
  // will be unoptimized and forced into software.  See the DirectX documentation for more information
  // about using the shader debugger.
  dwShaderFlags := 0;
  {$IFDEF DEBUG_VS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
  {$ENDIF}
  {$IFDEF DEBUG_PS}
    dwShaderFlags := dwShaderFlags or D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
  {$ENDIF}

  // Load effect file containing the vertex shaders
  Result := DXUtil_FindMediaFileCch(strPath, MAX_PATH, 'LightingVS.fx');
  if FAILED(Result) then Exit;

  Result := D3DXCreateEffectFromFile(m_pd3dDevice, strPath, nil, nil,
                                     dwShaderFlags, nil, m_pEffect, nil);
  if FAILED(Result) then Exit;

  hTech := m_pEffect.GetTechnique(m_iCurTechnique);
  m_pEffect.GetTechniqueDesc(hTech, TechDesc);
  m_pCurTechniqueName := TechDesc.Name;

  m_pEffect.SetTechnique(hTech);
  m_bTechniqueValid := SUCCEEDED(m_pEffect.ValidateTechnique(hTech));

  ZeroMemory(@m_light0, SizeOf(m_light0));
  m_light0._Type := D3DLIGHT_DIRECTIONAL;
  m_light0.Direction := D3DXVector3(0.3, -0.5, 0.2);
  m_light0.Diffuse.r := 0.25;
  m_light0.Diffuse.g := 0.25;
  m_light0.Diffuse.b := 0.25;

  ZeroMemory(@m_light1, SizeOf(m_light1));
  m_light1._Type := D3DLIGHT_DIRECTIONAL;
  m_light1.Direction := D3DXVector3(0.5, -0.5, 0.5);
  m_light1.Diffuse.r := 1.0;
  m_light1.Diffuse.g := 1.0;
  m_light1.Diffuse.b := 1.0;

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restores scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  mtrl: TD3DMaterial9;
  matIdentity: TD3DXMatrixA16;

  vFromPt: TD3DXVector3;
  vLookatPt: TD3DXVector3;
  vUpVec: TD3DXVector3;
  fAspect: Single;
begin
  Result:= BuildWallMesh;
  if FAILED(Result) then Exit;

  Result:= E_FAIL;

  // Create sphere and cone meshes to represent the lights
  if FAILED(D3DXCreateSphere(m_pd3dDevice, 0.25, 20, 20, m_pSphereMesh, nil))
  then Exit;

  if FAILED(D3DXCreateCylinder(m_pd3dDevice, 0.0, 0.25, 0.5, 20, 20, m_pConeMesh, nil)) 
  then Exit;

  // Set up a material
  ZeroMemory(@mtrl, SizeOf(TD3DMaterial9));
  mtrl.Diffuse.r := 1.0; mtrl.Ambient.r := 1.0;
  mtrl.Diffuse.g := 1.0; mtrl.Ambient.g := 1.0;
  mtrl.Diffuse.b := 1.0; mtrl.Ambient.b := 1.0;
  mtrl.Diffuse.a := 1.0; mtrl.Ambient.a := 1.0;
  m_pd3dDevice.SetMaterial(mtrl);

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);

  // Set the world matrix
  D3DXMatrixIdentity(matIdentity);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, matIdentity);

  // Set the view matrix.
  vFromPt   := D3DXVector3(-10, 10, -10);
  vLookatPt := D3DXVector3(0.0, 0.0, 0.0);
  vUpVec    := D3DXVector3(0.0, 1.0, 0.0);
  D3DXMatrixLookAtLH(m_matView, vFromPt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);

  // Set the projection matrix
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(m_matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProj);

  // Turn on lighting.
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  // Enable ambient lighting to a dim, grey light, so objects that
  // are not lit by the other lights are not completely black
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,
    D3DCOLOR_COLORVALUE(0.25, 0.25, 0.25, 1.0));

  // Set light #0 to be a simple, faint grey directional light so
  // the walls and floor are slightly different shades of grey
  m_pd3dDevice.SetLight(0, m_light0);

  // Set light #1 to be a simple, bright directional light to use
  // on the mesh representing light #2
  m_pd3dDevice.SetLight(1, m_light1);

  // Light #2 will be the light used to light the floor and walls.  It will
  // be set up in FrameMove() since it changes every frame.

  // Restore the font
  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  if (m_pEffect <> nil) then m_pEffect.OnLostDevice;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: BuildWallMesh()
// Desc: Builds the wall mesh based on m_m and m_n.
//-----------------------------------------------------------------------------
function CMyD3DApplication.BuildWallMesh: HRESULT;
var
  v: PMyVertexArray;
  pWallMeshTemp: ID3DXMesh;
  pdwAdjacency: PDWORD;
  dX, dZ: Single;
  k, z, x: LongWord;
  pIndex: PWord;
  iIndex: Word;
label
  _End;
begin
  v := nil;
  pWallMeshTemp := nil;
  pdwAdjacency := nil;

  SAFE_RELEASE(m_pWallMesh);

  // Create a square grid m_n*m_m for rendering the wall
  Result := D3DXCreateMeshFVF(m_nTriangles, m_nTriangles * 3,
               D3DXMESH_SYSTEMMEM, TMyVertex_FVF, m_pd3dDevice, pWallMeshTemp);
  if FAILED(Result) then goto _End;

  // Fill in the grid vertex data
  Result:= pWallMeshTemp.LockVertexBuffer(0, Pointer(v));
  if FAILED(Result) then goto _End;

  dX := 1.0/(m_n-1);
  dZ := 1.0/(m_m-1);
  k := 0;
  for z := 0 to (m_m-2) do
  begin
    for x := 0 to (m_n-2) do
    begin
      v[k].p := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * x*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * x*dX, 0.0, 10 * z*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * (z+1)*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
      v[k].p := D3DXVector3(10 * (x+1)*dX, 0.0, 10 * z*dZ);
      v[k].n := D3DXVector3(0.0, 1.0, 0.0);
      Inc(k);
    end;
  end;
  pWallMeshTemp.UnlockVertexBuffer;

  // Fill in index data
  Result:= pWallMeshTemp.LockIndexBuffer(0, Pointer(pIndex));
  if FAILED(Result) then goto _End;

  for iIndex := 0 to m_nTriangles * 3 - 1 do
  begin
    pIndex^:= iIndex;
    Inc(pIndex);
  end;
  pWallMeshTemp.UnlockIndexBuffer;

  // Eliminate redundant vertices
  try
    GetMem(pdwAdjacency, SizeOf(DWORD) * 3 * m_nTriangles);
  except
    Result := E_OUTOFMEMORY;
  end;
  if FAILED(Result) then goto _End;

  Result := D3DXWeldVertices(pWallMeshTemp, D3DXWELDEPSILONS_WELDALL, nil, nil, pdwAdjacency, nil, nil);
  if FAILED(Result) then
  begin
    goto _End;
  end;

  // Optimize the mesh
  Result := pWallMeshTemp.OptimizeInplace(D3DXMESHOPT_COMPACT or D3DXMESHOPT_VERTEXCACHE,
             pdwAdjacency, nil, nil, nil);
  if FAILED(Result) then goto _End;

  // Copy the mesh into fast write-only memory
  Result := pWallMeshTemp.CloneMeshFVF(D3DXMESH_WRITEONLY, TMyVertex_FVF,
              m_pd3dDevice, m_pWallMesh);
  if FAILED(Result) then goto _End;

_End:
  SAFE_RELEASE(pWallMeshTemp);
  FreeMem(pdwAdjacency); // SAFE_DELETE_ARRAY(pdwAdjacency);
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  dwCaps: DWORD;
  i: Integer;
  x, y, z: Single;
  lightType: TD3DLightType;
begin
  ZeroMemory(@m_light, SizeOf(m_light));

  if (m_CurrentLightType = TD3DLightType(0)) then
    // Rotate through the various light types
    lightType := TD3DLightType(1 + (Trunc(m_fTime) div 5) mod 3)
  else
    lightType := m_CurrentLightType;

  // Make sure the light type is supported by the device.  If
  // D3DVTXPCAPS_POSITIONALLIGHTS is not set, the device does not support
  // point or spot lights, so change light #2's type to a directional light.
  dwCaps := m_d3dCaps.VertexProcessingCaps;
  if (0 = dwCaps and D3DVTXPCAPS_POSITIONALLIGHTS) and
     (not m_bUseVertexShaders) then
  begin
    if (lightType = D3DLIGHT_POINT) or (lightType = D3DLIGHT_SPOT)
    then lightType := D3DLIGHT_DIRECTIONAL;
  end;

  for i := 0 to m_dwNumLights - 1 do
  begin
    m_light[i]._Type := lightType;

    // Values for the light position, direction, and color
    x := Sin(m_fTime*2.000*(i+1.5));
    y := Sin(m_fTime*2.246*(i+1.5));
    z := Sin(m_fTime*2.640*(i+1.5));

    m_light[i].Diffuse.r  := 0.5 + 0.5 * x;
    m_light[i].Diffuse.g  := 0.5 + 0.5 * y;
    m_light[i].Diffuse.b  := 0.5 + 0.5 * z;
    m_light[i].Range      := 100.0;

    case lightType of
      D3DLIGHT_POINT:
      begin
        // m_light[i].Position     = 4.5f * D3DXVECTOR3( x, y, z );
        D3DXVec3Scale(m_light[i].Position, D3DXVector3(x, y, z), 4.5);
        m_light[i].Attenuation1 := 0.4;
      end;

      D3DLIGHT_DIRECTIONAL:
        m_light[i].Direction    := D3DXVector3(x, y, z);

      D3DLIGHT_SPOT:
      begin
        // m_light[i].Position     := 2.0 * D3DXVector3(x, y, z);
        D3DXVec3Scale(m_light[i].Position, D3DXVector3(x, y, z), 2.0);
        m_light[i].Direction    := D3DXVector3(x, y, z);
        m_light[i].Theta        := 0.5;
        m_light[i].Phi          := 1.0;
        m_light[i].Falloff      := 1.0;
        m_light[i].Attenuation0 := 1.0;
      end;
    end;
    m_pd3dDevice.SetLight(2+i, m_light[i]);
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetMatrices()
// Desc: Called once a world or view matrix is changed to update vertex shader
//       constants
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetMatrices(const pMatWorld, pMatViewProj: TD3DXMatrix);
var
  m: TD3DXMatrix;
begin
  D3DXMatrixMultiplyTranspose(m, pMatWorld, pMatViewProj);
  m_pd3dDevice.SetVertexShaderConstantF(100, PSingle(@m), 4);
  D3DXMatrixMultiplyTranspose(m, pMatWorld, m_matView);
  m_pd3dDevice.SetVertexShaderConstantF(104, PSingle(@m), 4);
  D3DXMatrixMultiply(m, pMatWorld, m_matView);
  D3DXMatrixInverse(m, nil, m);
  m_pd3dDevice.SetVertexShaderConstantF(108, PSingle(@m), 4);
end;


//-----------------------------------------------------------------------------
// Name: SetLight()
// Desc: Called to set light parameters as vertex shader constants
//       constants
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetLight(LightIndex: DWORD; const light: TD3DLight9);
var
  direction: TD3DXVector4;
  position: TD3DXVector4;
  tmp: array[0..8] of Single;
begin
  m_pd3dDevice.SetVertexShaderConstantF(LightIndex*NUM_VECTORS_PER_LIGHT + 5, PSingle(@light.Ambient),  1);
  m_pd3dDevice.SetVertexShaderConstantF(LightIndex*NUM_VECTORS_PER_LIGHT + 6, PSingle(@light.Diffuse),  1);
  m_pd3dDevice.SetVertexShaderConstantF(LightIndex*NUM_VECTORS_PER_LIGHT + 7, PSingle(@light.Specular), 1);
  // Transform light direction and position to the camera space
  D3DXVec3TransformNormal(PD3DXVector3(@direction)^, PD3DXVector3(@light.Direction)^, m_matView);
  D3DXVec3Normalize(PD3DXVector3(@direction)^, PD3DXVector3(@direction)^);
  m_pd3dDevice.SetVertexShaderConstantF(LightIndex*NUM_VECTORS_PER_LIGHT + 8, PSingle(@direction), 1);
  D3DXVec3TransformCoord(PD3DXVector3(@position)^, PD3DXVector3(@light.Position)^, m_matView);
  m_pd3dDevice.SetVertexShaderConstantF(LightIndex * NUM_VECTORS_PER_LIGHT + 9, PSingle(@position), 1);

  tmp[0]:= light.Attenuation0;
  tmp[1]:= light.Attenuation1;
  tmp[2]:= light.Attenuation2;
  tmp[3]:= Single(light._Type);
  tmp[4]:= light.Falloff;
  tmp[5]:= -Cos(light.Phi/2.0);
  tmp[6]:= 1.0/(Cos(light.Theta/2.0) - Cos(light.Phi/2.0));
  tmp[7]:= 0;
  m_pd3dDevice.SetVertexShaderConstantF(LightIndex*NUM_VECTORS_PER_LIGHT + 10, @tmp, SizeOf(tmp) div ((SizeOf(Single)*4)));
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
var
  matWorld, matTrans, matRotate: TD3DXMatrixA16;
  numPasses: LongWord;
  matViewProj: TD3DXMatrix;
  i: Integer;
  mtrl: TD3DMaterial9;
  tmp: TD3DXVector4; // array[0..3] of Single;

  dwNumDirectionalLights: DWORD;   // m_light0 is directional
  dwNumSpotLights: DWORD;
  dwNumPointLights: DWORD;

  dwCurrentDirectionalLightIndex: DWORD;
  dwCurrentSpotLightIndex: DWORD;
  dwCurrentPointLightIndex: DWORD;

  i0: array[0..11] of Integer;
  iPass: Integer;
  vecFrom, vecAt, vecUp: TD3DXVector3;
  matWorldInv: TD3DXMatrixA16;

  strLight: PChar;
  strInfo: array[0..299] of Char;
const
  c0: array[0..7] of Single = (
    0,1,2,3,                         // Useful constants
    0.25, 0.25, 0.25, 1              // Global ambient factor
  );

begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Only render the walls and lights if using the FF pipeline or
    // a valid effect technique
    if (m_bTechniqueValid or not m_bUseVertexShaders) then
    begin
      if m_bWireframe
      then m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_WIREFRAME)
      else m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

      if m_bUseVertexShaders then
      begin
        // Set vertex shader
        m_pEffect._Begin(@numPasses, 0);

        // Precompute m_view * m_projection
        D3DXMatrixMultiply(matViewProj, m_matView, m_matProj);

        // Set up material as vertex shader constants
        ZeroMemory(@mtrl, SizeOf(TD3DMaterial9));
        mtrl.Diffuse.r := 1.0; mtrl.Ambient.r := 1.0;
        mtrl.Diffuse.g := 1.0; mtrl.Ambient.g := 1.0;
        mtrl.Diffuse.b := 1.0; mtrl.Ambient.b := 1.0;
        mtrl.Diffuse.a := 1.0; mtrl.Ambient.a := 1.0;
        m_pd3dDevice.SetVertexShaderConstantF(0, PSingle(@mtrl.Ambient),  1);
        m_pd3dDevice.SetVertexShaderConstantF(1, PSingle(@mtrl.Emissive), 1);
        m_pd3dDevice.SetVertexShaderConstantF(2, PSingle(@mtrl.Diffuse),  1);
        m_pd3dDevice.SetVertexShaderConstantF(3, PSingle(@mtrl.Specular), 1);
        tmp:= D3DXVector4(mtrl.Power, 0, 0, 0);
        m_pd3dDevice.SetVertexShaderConstantF(4, @tmp, 1);

        SetLight(0, m_light0);

        dwNumDirectionalLights := 1;   // m_light0 is directional
        dwNumSpotLights := 0;
        dwNumPointLights := 0;
        // Go through the lights and find number of each light type
        for i:=0 to m_dwNumLights - 1 do
        begin
          case m_light[i]._Type of
            D3DLIGHT_DIRECTIONAL: Inc(dwNumDirectionalLights);
            D3DLIGHT_SPOT:        Inc(dwNumSpotLights);
            D3DLIGHT_POINT:       Inc(dwNumPointLights);
          end;
        end;

        dwCurrentDirectionalLightIndex := 1;
        dwCurrentSpotLightIndex := 0;
        dwCurrentPointLightIndex := 0;
        // Set light params as vertex shader constants
        for i:=0 to m_dwNumLights - 1 do
        begin
          case m_light[i]._Type of
            D3DLIGHT_DIRECTIONAL:
            begin
              SetLight(dwCurrentDirectionalLightIndex, m_light[i]);
              Inc(dwCurrentDirectionalLightIndex);
            end;

            D3DLIGHT_SPOT:
            begin
              SetLight(dwNumDirectionalLights + dwCurrentSpotLightIndex, m_light[i]);
              Inc(dwCurrentSpotLightIndex);
            end;
            D3DLIGHT_POINT:
            begin
              SetLight(dwNumDirectionalLights + dwNumSpotLights + dwCurrentPointLightIndex, m_light[i]);
              Inc(dwCurrentPointLightIndex);
            end;
          end;
        end;

        // Set various constants
        (*
        float c0[] = {
            0,1,2,3,                            // Useful constants
            0.25f, 0.25f, 0.25f, 1              // Global ambient factor
        };
        *)
        m_pd3dDevice.SetVertexShaderConstantF(112, @c0, SizeOf(c0) div (SizeOf(Single)*4));

        // Loop parameters for each light type
        (*
              int i0[] = {
                  dwNumDirectionalLights , 0, NUM_VECTORS_PER_LIGHT, 0,
                  dwNumSpotLights,  dwNumDirectionalLights * NUM_VECTORS_PER_LIGHT, NUM_VECTORS_PER_LIGHT, 0,
                  dwNumPointLights, (dwNumDirectionalLights + dwNumSpotLights) * NUM_VECTORS_PER_LIGHT, NUM_VECTORS_PER_LIGHT, 0
              };
        *)
        i0[ 0] := dwNumDirectionalLights;
        i0[ 1] := 0;
        i0[ 2] := NUM_VECTORS_PER_LIGHT;
        i0[ 3] := 0;
        i0[ 4] := dwNumSpotLights;
        i0[ 5] := dwNumDirectionalLights * NUM_VECTORS_PER_LIGHT;
        i0[ 6] := NUM_VECTORS_PER_LIGHT;
        i0[ 7] := 0;
        i0[ 8] := dwNumPointLights;
        i0[ 9] := (dwNumDirectionalLights + dwNumSpotLights) * NUM_VECTORS_PER_LIGHT;
        i0[10] := NUM_VECTORS_PER_LIGHT;
        i0[11] := 0;

        m_pd3dDevice.SetVertexShaderConstantI(0, @i0, SizeOf(i0) div (SizeOf(Integer)*4));
      end else
      begin
        numPasses := 1;
        // Turn on light #0 and #2, and turn off light #1
        m_pd3dDevice.LightEnable(0, True);
        m_pd3dDevice.LightEnable(1, False);
        for i:=0 to m_dwNumLights - 1 do
          m_pd3dDevice.LightEnable(2 + i, True);
      end;

      for iPass := 0 to numPasses - 1 do
      begin
        if m_bUseVertexShaders then m_pEffect.Pass(iPass);

        // Draw the floor
        D3DXMatrixTranslation(matTrans, -5.0, -5.0, -5.0);
        D3DXMatrixRotationZ(matRotate, 0.0);
        D3DXMatrixMultiply(matWorld, matRotate, matTrans);
        if m_bUseVertexShaders
          then SetMatrices(matWorld, matViewProj)
          else m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
        m_pWallMesh.DrawSubset(0);

        // Draw the back wall
        D3DXMatrixTranslation(matTrans, 5.0, -5.0, -5.0);
        D3DXMatrixRotationZ(matRotate, D3DX_PI/2);
        D3DXMatrixMultiply(matWorld, matRotate, matTrans);
        if m_bUseVertexShaders
          then SetMatrices(matWorld, matViewProj)
          else m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
        m_pWallMesh.DrawSubset(0);

        // Draw the side wall
        D3DXMatrixTranslation(matTrans, -5.0, -5.0, 5.0);
        D3DXMatrixRotationX(matRotate,  -D3DX_PI/2);
        D3DXMatrixMultiply(matWorld, matRotate, matTrans);
        if m_bUseVertexShaders
          then SetMatrices(matWorld, matViewProj)
          else m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
        m_pWallMesh.DrawSubset(0);
      end;
      if m_bUseVertexShaders then m_pEffect._End;

      if m_bUseVertexShaders then
      begin
        SetLight(0, m_light1);
        (*
        int i0[] = {1, 0, NUM_VECTORS_PER_LIGHT, 0,
                    0,0,0,0,
                    0,0,0,0};
        *)
        FillChar(i0, 12*3, 0);
        i0[ 0] := 1;
        i0[ 2] := NUM_VECTORS_PER_LIGHT;

        m_pd3dDevice.SetVertexShaderConstantI(0, @i0, SizeOf(i0) div (SizeOf(Integer)*4));
      end else
      begin
        // Turn on light #1, and turn off light #0 and #2
        m_pd3dDevice.LightEnable(0, False);
        m_pd3dDevice.LightEnable(1, True);
        for i:=0 to m_dwNumLights - 1 do
          m_pd3dDevice.LightEnable( 2 + i, False);
      end;

      m_pd3dDevice.SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);

      // Draw the mesh representing the light
      for i:= 0 to m_dwNumLights - 1 do
      begin
        if (m_light[i]._Type = D3DLIGHT_POINT) then
        begin
          // Just position the point light -- no need to orient it
          D3DXMatrixTranslation(matWorld, m_light[i].Position.x,
            m_light[i].Position.y, m_light[i].Position.z);

          if m_bUseVertexShaders then
          begin
            SetMatrices(matWorld, matViewProj);
            m_pEffect._Begin(@numPasses, 0);
            for iPass := 0 to numPasses - 1 do
            begin
              m_pEffect.Pass(iPass);
              m_pSphereMesh.DrawSubset(0);
            end;
            m_pEffect._End;
          end else
          begin
            m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
            m_pSphereMesh.DrawSubset(0);
          end;

        end else
        begin
          // Position the light and point it in the light's direction
          vecFrom:= D3DXVector3(m_light[i].Position.x, m_light[i].Position.y, m_light[i].Position.z);
          vecAt:= D3DXVector3(m_light[i].Position.x + m_light[i].Direction.x,
                    m_light[i].Position.y + m_light[i].Direction.y,
                    m_light[i].Position.z + m_light[i].Direction.z);
          vecUp:= D3DXVector3(0, 1, 0);
          D3DXMatrixLookAtLH(matWorldInv, vecFrom, vecAt, vecUp);
          D3DXMatrixInverse(matWorld, nil, matWorldInv);
          if m_bUseVertexShaders then
          begin
            SetMatrices(matWorld, matViewProj);
            m_pEffect._Begin(@numPasses, 0);
            for iPass := 0 to numPasses - 1 do
            begin
              m_pEffect.Pass(iPass);
              m_pConeMesh.DrawSubset(0);
            end;
            m_pEffect._End;
          end else
          begin
            m_pd3dDevice.SetTransform(D3DTS_WORLD, matWorld);
            m_pConeMesh.DrawSubset(0);
          end;
        end;
      end;
    end;
    // Output statistics
    m_pFont.DrawText( 2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText( 2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    case m_light[0]._Type of
      D3DLIGHT_POINT: strLight := 'Point Light';
      D3DLIGHT_SPOT : strLight := 'Spot Light';
     else
       strLight := 'Directional Light';
    end;

    if (m_dwNumLights <> 0) then
    begin
      StrFmt(strInfo, '%d %s, %dx%d vertices in wall/floor mesh', [m_dwNumLights, strLight, m_n, m_m]);
      m_pFontSmall.DrawText( 2, 40, D3DCOLOR_ARGB(255,255,255,255), strInfo);
    end;

    if m_bUseVertexShaders then
    begin
      StrFmt(strInfo, 'Using vertex shader: %s', [m_pCurTechniqueName]);
      if (not m_bTechniqueValid) then
        StrCat(strInfo, ' (Warning: validation failed)');
      m_pFontSmall.DrawText( 2, 60, D3DCOLOR_ARGB(255,255,255,255), strInfo);
    end else
      m_pFontSmall.DrawText( 2, 60, D3DCOLOR_ARGB(255,255,255,255), 'Using fixed function');

    if m_bShowHelp then
    begin
      m_pFontSmall.DrawText( 2, 80, D3DCOLOR_ARGB(255,255,255,255),
        'Use arrow keys to change mesh density');
      m_pFontSmall.DrawText( 2, 100, D3DCOLOR_ARGB(255,255,255,255),
        'Press W to toggle wireframe mode');
      m_pFontSmall.DrawText( 2, 120, D3DCOLOR_ARGB(255,255,255,255),
        'Press S to toggle between fixed function and programmable vertex pipeline');
      m_pFontSmall.DrawText( 2, 140, D3DCOLOR_ARGB(255,255,255,255),
        'Press D to decrement number of lights');
      m_pFontSmall.DrawText( 2, 160, D3DCOLOR_ARGB(255,255,255,255),
        'Press I to increment number of lights');
      m_pFontSmall.DrawText( 2, 180, D3DCOLOR_ARGB(255,255,255,255),
        'Press T to change light type');
      m_pFontSmall.DrawText( 2, 200, D3DCOLOR_ARGB(255,255,255,255),
        'Press V to change vertex shader');
    end else
    begin
      m_pFontSmall.DrawText(2, 80, D3DCOLOR_ARGB(255,255,255,255), 'Press F1 for help');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Invalidates device objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  SAFE_RELEASE(m_pWallMesh);
  SAFE_RELEASE(m_pSphereMesh);
  SAFE_RELEASE(m_pConeMesh);
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;

  if (m_pEffect <> nil) then m_pEffect.OnLostDevice;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  SAFE_RELEASE(m_pEffect);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  // Cleanup D3D font
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFontSmall);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: LongWord;
  wParam: WPARAM; lParam: LPARAM): LRESULT;
var
  EffectDesc: TD3DXEffectDesc;
  TechDesc: TD3DXTechniqueDesc;
  hTech: TD3DXHandle;
begin
  case uMsg of
    WM_COMMAND:
    case LOWORD(wParam) of
      IDM_TOGGLEHELP:       m_bShowHelp := not m_bShowHelp;
      IDM_TOGGLEWIREFRAME:  m_bWireframe := not m_bWireframe;
      IDM_TOGGLESHADERS:    m_bUseVertexShaders := not m_bUseVertexShaders;
      IDM_TOGGLEVERTEXSHADERS:
      begin
        m_pEffect.GetDesc(EffectDesc);
        Inc(m_iCurTechnique);
        if (m_iCurTechnique >= EffectDesc.Techniques) then m_iCurTechnique := 0;
        hTech := m_pEffect.GetTechnique(m_iCurTechnique);
        m_pEffect.GetTechniqueDesc(hTech, TechDesc);
        m_pCurTechniqueName := TechDesc.Name;
        m_pEffect.SetTechnique(hTech);
        m_bTechniqueValid := SUCCEEDED(m_pEffect.ValidateTechnique(hTech));
      end;

      IDM_INCNUMLIGHTS:
        if (m_dwNumLights < MAXLIGHTS) then Inc(m_dwNumLights);
      IDM_DECNUMLIGHTS:
        if (m_dwNumLights > 0) then Dec(m_dwNumLights);
      IDM_CHANGELIGHTTYPE:
      begin
        m_CurrentLightType := TD3DLightType(DWORD(m_CurrentLightType) + 1);
        if (Ord(m_CurrentLightType) > 3) then
            m_CurrentLightType := TD3DLightType(0);
      end;

      IDM_INCREASERESOLUTION:
        if (m_n < 64) and (m_m < 64) then
        begin
          m_n := m_n *2;
          m_m := m_m *2;
          m_nTriangles := (m_n-1) * (m_m-1) * 2;
          BuildWallMesh;
        end;
      IDM_DECREASERESOLUTION:
        if (m_n > 2) and (m_m > 2) then
        begin
          m_n := m_n div 2;
          m_m := m_m div 2;
          m_nTriangles := (m_n-1) * (m_m-1) * 2;
          BuildWallMesh;
        end;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

